#!/usr/bin/gawk -f
##################################################################
# The Pthread BE generator
#
#Copyright 2008 Pjotr Kourzanov
#
#This file is part of LIME.
#
#LIME is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License version 2
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#LIME is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with LIME.  If not, see <http://www.gnu.org/licenses/>.
##################################################################
BEGIN { if (!OP) OP="dump"
       	if (!STDERR) STDERR="/dev/stderr"
	state=""  #top
}

func handle(tag,body) { return handle_gxf(tag,handle_types(tag,body)) }

END {
    merge_types()

    if (OP=="main.c") for (f in files) if (f ~ /.gxf$/) genmain_c(f)
    if (OP=="shell.h") for (f in files) if (f ~ /.gxf$/) genshell_h(f)
    if (OP=="shell.c") for (f in files) if (f ~ /.gxf$/) genshell_c(f)
}

func comp_attr(n) { return n "_attr" }
func comp_affn(n) { return n "_affn" }
func comp_args(n) { return n "_args" }
func comp_parm(n) { return n "_parm" }

func fifo_attr(n) { return "fifo" n "_attr" }
func fifo_id(n) { return "fifo" n }
func fifo_name(fn,fp,tn,tp) { return "/" fn "_" fp "__" tn "_" tp }

func genmain_c(f, i,pr,fid,n,c,j,sn,pv,k,fs,isn,isp) {
    pr=gensub(/\.[a-z]+$/,"","g",f)
    print "#include <stdlib.h>"
    print "#include <pthread.h>"
    print "#include <mqueue.h>"
    print "#include <debugging.h>"
    print "#include <assert.h>"
    if (TRACE) while (getline < TRACE) print $0
    close(TRACE)
    for (i=1; i<=nnodes[f]; i++) {
    	n=nodes[f,i]
	# don't include header for unconnected nodes
	if (!fanout[f,n] && !fanin[f,n]) continue
	c=gensub(/\$[a-z]+$/,"","g",n)
	if (!is_node_type(f,n,"extern")) {
		print "#include","\"" c "-shell.h\""
	}
    }
    print "int main(int argc, char* argv[]) {"
    print space(2) "void *_;"
    # task id variables
    print space(2) "// task config"
    for (i=1; i<=nnodes[f]; i++) {
	n=nodes[f,i]
	# {con,de}structors do not have a task id
	if (is_structor(f,n)) continue
	# don't create a task for unconnected nodes
	if (!is_connected(f,n)) continue
	if (node_type[f,n]=="extern") continue
	if (!is_fullyconnected(f,n)) fail(f ":" n " not fully connected")

	print space(2) "pthread_t",comp_tid(n) ";"
	print space(2) "pthread_attr_t",comp_attr(n) ";"
	printf space(2) "struct %s_args %s = { .running=%s_RUNNING };\n",
		n,comp_args(n),n
    }
    # fifo attribute variables
    print space(2) "// fifo config"
    for (i=1; i<=nedges[f]; i++) {
	split(edge_from[f,i],fr,SUBSEP); fn=fr[1]; fp=fr[2]
	split(edge_to[f,i],to,SUBSEP); tn=to[1]; tp=to[2]
	#if (is_state(f,fn,"",fp)) continue
	if (!is_fifo(f,fn,"",fp)) continue
	# XXX: just take a max here
	fs="max(" eval(portsize(f,fn,fp)) "," eval(portsize(f,tn,tp)) ")"
	# include delay initializer is present
	if ((f,tn,tp) in delayport_idx)
		fs="max(" fs "," eval(portsize(f,init_node[f,tn,tp],init_port[f,tn,tp])) ")"
	printf space(2) "struct mq_attr %s = {.mq_msgsize=%s, .mq_maxmsg=%s };\n",
		fifo_attr(i),unittypesize(f,fn,fp),fs
	print space(2) "mqd_t",fifo_id(i) ";"
    }
    print "#if TRACING==1"
    print space(2) "lprintf(\"loading introspect %s...\\n\",argv[0]);"
    print space(2) "intro=introspect_init(argv[0]);"
    print space(2) "assert(intro);"
    print "#endif"

    print space(2) "lprintf(\"Hello POSIX\\n\");"
    print space(2) "// fifo creation"
    for (i=1; i<=nedges[f]; i++) {
	split(edge_from[f,i],fr,SUBSEP); fn=fr[1]; fp=fr[2]
	split(edge_to[f,i],to,SUBSEP); tn=to[1]; tp=to[2]
	#if (is_state(f,fn,"",fp)) continue
	if (!is_fifo(f,fn,"",fp)) continue
	printf space(2) "sys_ret(%s=mq_open(\"%s\",O_CREAT | O_WRONLY | O_EXCL, 0600, &%s));\n",
		fifo_id(i),fifo_name(fn,fp,tn,tp),fifo_attr(i)
	if (node_type[f,fn]!="extern")
	printf space(2) "sys_ret(%s.outq[%s]=mq_open(\"%s\",O_WRONLY));\n",
		comp_args(fn),
		fifoidx(f,fn,fp),
		fifo_name(fn,fp,tn,tp)
	if (node_type[f,tn]!="extern")
	printf space(2) "sys_ret(%s.inpq[%s]=mq_open(\"%s\",O_RDONLY));\n",
		comp_args(tn),
		fifoidx(f,tn,tp),
		fifo_name(fn,fp,tn,tp)
	printf space(2) "sys_ret(mq_unlink(\"%s\"));\n",
		fifo_name(fn,fp,tn,tp)
    }
    # state initialization, for constructors that are linked to the NM
    print space(2) "// state initialization"
    for (i=1; i<=nstatefuls[f]; i++) {
	sn=statefuls[f,i]
	for (j=1; j<=nr_states(f,sn); j++) 
	{
	    p=stateports[f,sn,"read",j]
	    c=init_node[f,sn,p]
	    for (k=1; k<=nnodes[c]; k++) {
		n=nodes[c,k]
	    	args=""
	    	for (z=1; z<=nports[c,n]; z++) {
	  	  pv=ports[c,n,z]
	  	  args=append(args,
	      		(port_const[c,n,pv] ? "(const " port_type[c,n,pv] "*)" : "")\
	      		dump_portvar(f,sn,n,p, comp_args(sn) ".","."), ", ")
	    	}
	    	print space(2) scpe(c,n) "(" args ");"
	    	for (z=1; z<=noutports[c,n]; z++) {
	    	  pv=outports[c,n,z]
	     	  printf space(2) "%s_state=%s_BUSY;\n",
			 dump_portvar(f,sn,n,p,comp_args(sn) ".","."),sn
	    	}
	    }
	}
    }
    # delay handling, for nodes that are linked to the NM
    print space(2) "// delay handling"
    for (i=1; i<=ndelayed[f]; i++) {
	sn=delayed[f,i]
	print space(2) "{"
	for (j=1; j<=nr_delays(f,sn); j++) 
	{
	    p=delayports[f,sn,"read",j]
	    isn=init_node[f,sn,p]
	    isp=init_port[f,sn,p]
	    print space(4) get_portdef(f,isn,isp) ";"
	    c=peer_node[f,sn,p]
	    for (k=1; k<=nnodes[isn]; k++) {
		n=nodes[isn,k]
	    	args=""
	    	for (z=1; z<=nports[isn,n]; z++) {
	  	  pv=ports[isn,n,z]
	  	  args=append(args,
	      		(port_const[isn,n,pv] ? "(const " port_type[isn,n,pv] "*)" : "")\
	      		dump_portvar(f,isn,n,isp, sn "_",""), ", ")
	    	}
	    	print space(4) scpe(isn,n) "(" args ");"
	    	acquire_write(f,isn,n,isp,4,
	    		"exit(EXIT_FAILURE);",
			portsize(isn,n,isp),fifo_id(peer_edge[f,sn,p]))
	    	release_write(f,isn,n,isp,4,isp,
			unitportsize(isn,n,isp),fifo_id(peer_edge[f,sn,p]))
	    }
	}
	print space(2) "}"
    }
    print space(2) "// task setup"
    for (i=1; i<=nnodes[f]; i++) {
	n=nodes[f,i]
	if (is_structor(f,n)) continue
	# don't create a task for unconnected nodes
	if (!is_connected(f,n)) continue
	if (node_type[f,n]=="extern") continue
	if (!is_fullyconnected(f,n)) fail(f ":" n " not fully connected")

	printf space(2) "and_ret(pthread_attr_init(&%s));\n",
				comp_attr(n)
    }
    print space(2) "if (geteuid()==0) {"
    for (i=1; i<=nnodes[f]; i++) {
	n=nodes[f,i]
	# {con,de}structors do not have a task id
	if (is_structor(f,n)) continue
	# don't create a task for unconnected nodes
	if (!is_connected(f,n)) continue
	if (node_type[f,n]=="extern") continue
	if (!is_fullyconnected(f,n)) fail(f ":" n " not fully connected")

	printf space(4) "struct sched_param %s = {.sched_priority=%s};\n",
				comp_parm(n),"1"
if (do_affinity) {
	print space(4) "cpu_set_t",comp_affn(n) ";"
}
    }
    for (i=1; i<=nnodes[f]; i++) {
	n=nodes[f,i]
	if (is_structor(f,n)) continue
	# don't create a task for unconnected nodes
	if (!is_connected(f,n)) continue
	if (node_type[f,n]=="extern") continue
	if (!is_fullyconnected(f,n)) fail(f ":" n " not fully connected")

	printf space(4) "and_ret(pthread_attr_setschedpolicy(&%s, SCHED_FIFO));\n",
				comp_attr(n)
	printf space(4) "and_ret(pthread_attr_setschedparam(&%s, &%s));\n",
				comp_attr(n),comp_parm(n)

if (do_affinity) {
	printf space(4) "CPU_ZERO(&%s);\n",comp_affn(n)
	printf space(4) "CPU_SET(%s,&%s);\n",comp_core(n),comp_affn(n)
	printf space(4) "and_ret(pthread_attr_setaffinity_np(&%s, sizeof(%s), &%s));\n",
				comp_attr(n),comp_affn(n),comp_affn(n)
}
    }
    print space(2) "}"
    print space(2) "// task creation & startup"
    for (i=1; i<=nnodes[f]; i++) {
	n=nodes[f,i]
	if (is_structor(f,n)) continue
	# don't create a task for unconnected nodes
	if (!is_connected(f,n)) continue
	if (node_type[f,n]=="extern") continue
	if (!is_fullyconnected(f,n)) fail(f ":" n " not fully connected")

	printf space(2) "and_ret(pthread_create(&%s, &%s, (void* (*)(void*))%s, &%s));\n",
	       			comp_tid(n),comp_attr(n),comp_shell(n),comp_args(n)
    }
    print space(2) "sleep(5);"
    print space(2) "// task cleanup"
    for (i=1; i<=nnodes[f]; i++) {
	n=nodes[f,i]
	if (is_structor(f,n)) continue
	# don't create a task for unconnected nodes
	if (!is_connected(f,n)) continue
	if (node_type[f,n]=="extern") continue
	if (!is_fullyconnected(f,n)) fail(f ":" n " not fully connected")

	printf space(2) "%s.running=%s_STOPPED;\n",comp_args(n),n
	printf space(2) "if (1) and_ignore(pthread_cancel(%s));\n",
				comp_tid(n)
	printf space(2) "while (pthread_join(%s,&_));\n",
				comp_tid(n)
	printf space(2) "lprintf(\"%s returned %%p\\n\",_);\n",n
    }
    # state initialization, for constructors that are linked to the NM
    print space(2) "// state deinitialization"
    for (i=1; i<=nstatefuls[f]; i++) {
	sn=statefuls[f,i]
	for (j=1; j<=nr_states(f,sn); j++) 
	{
	    p=stateports[f,sn,"write",j]
	    c=deinit_node[f,sn,p]
	    for (k=1; k<=nnodes[c]; k++) {
		n=nodes[c,k]
	    	args=""
	    	for (z=1; z<=nports[c,n]; z++) {
	  	  pv=ports[c,n,z]
	  	  args=append(args,
	      		(port_const[c,n,pv] ? "(const " port_type[c,n,pv] "*)" : "")\
	      		dump_portvar(f,sn,n,p, comp_args(sn) ".","."), ", ")
	    	}
	    	print space(2) scpe(c,n) "(" args ");"
	    	for (z=1; z<=noutports[c,n]; z++) {
	    	  pv=outports[c,n,z]
	     	  printf space(2) "%s_state=%s_IDLE;\n",
			 dump_portvar(f,sn,n,p,comp_args(sn) ".","."),sn
	    	}
	    }
	}
    }
    print space(2) "// fifo cleanup"
    for (i=1; i<=nedges[f]; i++) {
	split(edge_from[f,i],fr,SUBSEP); fn=fr[1]; fp=fr[2]
	split(edge_to[f,i],to,SUBSEP); tn=to[1]; tp=to[2]
	#if (is_state(f,fn,"",fp)) continue
	if (edge_type[f,i]!="fifo") continue
	printf space(2) "sys_ret(mq_close(%s));\n",
		fifo_id(i)
	if (node_type[f,fn]!="extern")
	printf space(2) "sys_ret(mq_close(%s.outq[%s]));\n",
		comp_args(fn),
		fifoidx(f,fn,fp)
	if (node_type[f,tn]!="extern")
	printf space(2) "sys_ret(mq_close(%s.inpq[%s]));\n",
		comp_args(tn),
		fifoidx(f,tn,tp)
    }
    print "#if TRACING==1"
    print space(2) "lprint(\"unloading introspect...\");"
    print space(2) "introspect_deinit(intro);"
    print space(2) "intro=NULL;"
    print "#endif"

    print space(2) "return EXIT_SUCCESS;"
    print "}"
}

func genshell_h(f, i,n,fs,p,defined) {
    fs=PE "_h_defined" 
    print "#ifndef",fs
    print "#define",fs
    print "#include <pthread.h>"
    print "#include <mqueue.h>"
    delete included
    while (getline < IMPORTS) if (!included[$0]) {
    	print "#include \"" $0 ".h\""
	included[$0]=1
    }
    close(IMPORTS)
    n=PE
    if ((f,n) in node_idx); else fail("node " n " not found in " f)
    # {con,de}structors do not require a shell header, besides the
    # forward declaration for the actual call
    if (is_structor(f,n)) {
	dump_prototypes(f,n)
    	print "#endif//" fs
	return
    }

    if (nr_states(f,n)) {
      print comp_state(n),"{"
      # only generate state instance for incoming ports
      for (i=1; i<=nr_states(f,n); i++) {
    	p=stateports[f,n,"read",i]
	if (p in defined) continue
    	print space(2) "enum","_" n "_" p "_state {\n\t",
        	n "_IDLE = 0,\n\t",
      		n "_BUSY,\n\t",
      		n "_STOP\n",
    		"}", p "_state:2;" 
    	print space(2) get_portdef(f,n,p) ";"
	defined[p]=1
      }
      print "};"
    }
    if (nr_states(f,n))
      print "#define",n "_STATE_SIZE",unitsize(comp_state(n))
    else
      print "#define",n "_STATE_SIZE",0
    print "#define",n "_NR_INPUTS",ninfifos(f,n)
    print "#define",n "_NR_OUTPUTS",noutfifos(f,n)
    printf "struct %s_args {\n",PE

    if (nr_states(f,n))
    	printf space(2) "%s state;\n",comp_state(n)
    printf space(2) "enum { %s_STOPPED=0, %s_RUNNING=1 } running:1;\n",n,n
    printf space(2) "mqd_t inpq[%s_NR_INPUTS];\n",n
    printf space(2) "mqd_t outq[%s_NR_OUTPUTS];\n",n
    print "};"
    printf "extern void* %s(struct %s_args *);\n",comp_shell(n), PE
    print "#endif//" fs
}

func genshell_c(f, i,c,n,k) {
    print "#include <stdlib.h>","//exit"
    print "#include <debugging.h>","//lprintf"
    print "#include \"" PE "-shell.h\""
    # for all actors
    for (i=1; i<=nnodes[f]; i++) if ((c=nodes[f,i])==PE) {
    	# {con,de}structors do not require a shell
	if (is_structor(f,c)) continue

	if (nr_states(f,c))
		print "#include <assert.h>"

	dump_prototypes(f,c)

    	printf "void* %s(struct %s_args *args) {\n",comp_shell(c), PE

	print space(1) "int _;"
	print space(1) "sys_pexit(pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,&_),NULL);"
	print space(1) "sys_pexit(pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,&_),NULL);"

	print space(1) "lprintf(\"ENTERING [%u]",c "()\\n\"," c "_STATE_SIZE);"

	if (nr_states(f,c)) print space(2) "{",
		comp_state(c), 
		"*state=(" comp_state(c) "*)&(args->state);"

	delete declared; delete acquired

	printf space(1) "while (args->running==%s_RUNNING) {\n",c
	print space(2) "pthread_testcancel();"
	# for all sub-actors
	for (k=1; k<=nnodes[c]; k++)
	{
		n=nodes[c,k]
		print space(2) "{// cyclo-static actor:",n
		dump_ports(f,c,n)
		print space(2) "// ACQUIRES"
		do_acquires(f,c,n)
		do_reads(f,c,n)
		print space(2) "// DOWNCALL"
		do_call(f,c,n)
		print space(2) "// RELEASES"
		do_releases(f,c,n,k,"write")
		do_releases(f,c,n,k,"read")

		# strict CSDF requires sequential consumption
		# no port data sharing is allowed
		if (is_node_type(f,c,"strict")) {
			print space(2) "}"
			delete declared
			delete acquired
		}
	} 
	if (!is_node_type(f,c,"strict"))
		print space(2) space(nnodes[c],"}")
	if (nr_states(f,c)) print space(2) "}"
	print space(1) "}"
	print space(1) "return (void*)1;"
	print "}"
    }
}
func do_acquires(f,c,n, p,j,pv,tf,ti,a) {
    for (j=1; j<=nports[c,n];j++)
    {
	p=ports[c,n,j]
	if (p in acquired) continue
	# is this a shmem port?
	if (is_state(f,c,n,p)) {
	    # TODO: locking if state is shared?
	    if (!is_structor(c,n)) {
	    	pv=dump_portvar(f,c,n,p)
		printf space(2) "assert(%s_state==%s_BUSY);\n",pv,c
		printf space(2) "assert(%s);\n",pv
		acquired[p]=1
	    }
	    continue
	}
    	# is this is a dynamic port (of discriminated union type)?
	# nothing to do for output variant ports
	# (everything is at release time)
    	if (split(variants[port_type[c,n,p]],a,SUBSEP)==2)
	{ tf=a[1]; ti=a[2]

	  # non-deterministic selective multiport
	  if (port_const[c,n,p] && !is_node_type(c,n,"unfold"))
	  	read_selective_ndet(f,c,n,p,tf,ti)

	  # deterministic selective multiport
    	  if(port_const[c,n,p] && is_node_type(c,n,"unfold")) 
	    	read_selective_det(f,c,n,p,tf,ti)

	  continue
	}
	# if no - use it for activation (do an acquire)
	if (port_const[c,n,p]) {
		acquire_read(f,c,n,p,2,p,
			"continue;")
	} else acquire_write(f,c,n,p,2,
			"continue;",
			portsize(c,n,p))

	acquired[p]=1
    }
}
func do_reads(f,c,n, p,j) {
    for (j=1; j<=nports[c,n];j++)
    {
	p=ports[c,n,j]
	if (p in _read) continue

	if (is_state(f,c,n,p))
		continue

    	if (split(variants[port_type[c,n,p]],a,SUBSEP)==2)
	  continue

	# if no - use it for activation (do an acquire)
	if (port_const[c,n,p]) 
		do_read(f,c,n,p,2,p, unitportsize(c,n,p))

	_read[p]=1
    }
}
func read_selective_det(f,c,n,p,tf,ti, z,dp,t) {
     t=types[tf,ti]
     dp=p "_" typesels[tf,ti]
     # acquire selectors
     for (z=0; z<portsize(f,c,dp); z++) {
         acquire_read(f,c,0,dp,2,
			"&" variantsel(p,tf,ti,z),
			"continue;",
			portsize(f,c,dp))
         do_read(f,c,0,dp,2,
			"&" variantsel(p,tf,ti,z),
			unittypesize(f,c,dp))
     }
     # actually read data
     print space(2) "{ int i; for (i=0; i<" portsize(c,n,p) "; i++)",
                "switch (" variantsel(p,tf,ti,"i") ") {"
     for (z=1; z<=nports[tf,t]; z++) {
     	dp=ports[tf,t,z]
     	print space(4) "case",port_sel[tf,t,dp] ":"
	printf space(6) "lprintf(\"Selected %s\\n\");\n",port_sel[tf,t,dp]
     	acquire_read(f,c,0,dp,6,
			variantport(p,tf,ti,"i") "." dp,
			"continue;",
			portsize(f,c,dp))
     	do_read(f,c,0,dp,6,
			variantport(p,tf,ti,"i") "." dp,
			unitportsize(f,c,dp))
     	print space(6) "break;"
     }
     print space(2) "}}"
}
func read_selective_ndet(f,c,n,p,tf,ti, z,dp,t) {
    t=types[tf,ti]
    print space(2) "{ int i;","for (i=0; i<" portsize(c,n,p) "; i++)"
    print space(3) "{ int p;","for (p=" fifoidx(f,c,ports[tf,t,1]) ";",
    			"p<=" fifoidx(f,c,ports[tf,t,nports[tf,t]]) ";",
			"p++) {"
    for (z=1; z<=nports[tf,t]; z++) {
	dp=ports[tf,t,z]
    	print space(4) "if (p==" fifoidx(f,c,ports[tf,t,z]) ") {"
    	acquire_read(f,c,0,dp,6,
			variantport(p,tf,ti,"i") "." dp,
			"continue;",
			portsize(f,c,dp))
    	do_read(f,c,0,dp,6,
			variantport(p,tf,ti,"i") "." dp,
			unitportsize(f,c,dp))
	print space(6) variantsel(p,tf,ti,"i") "=" port_sel[tf,t,dp] ";"
	printf space(6) "lprintf(\"Selected %s\\n\");\n",port_sel[tf,t,dp]
    	print space(6) "break;"
	print space(4) "}"
    }
    print space(3) "}"
    print space(3) "if (p>" fifoidx(f,c,ports[tf,t,nports[tf,t]]) ")",
    			"continue;"
    print space(3) "}"
    print space(2) "}"
}

func do_call(f,c,n, p,args,cond,chng,j,pv) {
  for (j=1; j<=nports[c,n]; j++) {
    p=ports[c,n,j]
    pv=dump_portvar(f,c,n,p)
    if ((c,n,p) in cons_idx) {
	    cond=append(cond,pv "_state==" c "_IDLE","&&")
	    chng=append(chng,pv "_state=" c "_BUSY;","\n")
    }
    if ((c,n,p) in dstr_idx) {
	    cond=append(cond,pv "_state==" c "_STOP","&&")
	    chng=append(chng,pv "_state=" c "_IDLE;","\n")
    }
    if ((f,c,p) in fifo_idx && !port_const[f,c,p]) {
    	    noutfif[f,c,n]++
    }
    args=append(args,
		(port_const[c,n,p] ? "(const " port_type[c,n,p] "*)" : "")\
		pv, ", ")
  }
  if (cond && chng) {
	print space(2) "if (" cond ") {"
	print space(4) scpe(c,n) "(" args ");"
	print space(4) chng
	print space(2) "}"
	return
  }
  if (noutfif[f,c,n]==1 && node_ret[c,n]!="void") {
  	print space(2) node_ret[c,n] " ret=" scpe(c,n) "(" args ");"
	return
  }
  print space(2) scpe(c,n) "(" args ");"
}

func do_releases(f,c,n,k,dir, p,j,tf,ti,t,a,z,dp,gen) {
  for (j=1; j<=nports[c,n];j++) {
    p=ports[c,n,j]
    # is this a shmem port?
    if (is_state(f,c,n,p)) {
	# TODO: locking if state is shared?
	continue
    }
    # is this is a dynamic port (of discriminated union type)?
    if (split(variants[port_type[c,n,p]],a,SUBSEP)==2) 
    {	tf=a[1]; ti=a[2]; t=types[tf,ti]

	gen=( (dir=="read" && port_const[c,n,p] && nports[tf,t]) ||
	     (dir=="write" && !port_const[c,n,p] && nports[tf,t]) )

	# XXX - acquire selectors
	dp=p "_" typesels[tf,ti]
	if (!port_const[c,n,p] && dir=="write") for (z=0; z<portsize(f,c,dp); z++) {
		acquire_write(f,c,0,dp,2,
			"continue;",
			portsize(f,c,dp))
	}

	if (gen) print space(2) "{ int i; for (i=0; i<" portsize(c,n,p) "; i++)",
	               "switch (" variantsel(p,tf,ti,"i") ") {"

	for (z=1; z<=nports[tf,t]; z++) {
	    dp=ports[tf,t,z]
	    if (gen) print space(4) "case",port_sel[tf,t,dp] ":"
	    if (port_const[f,c,dp] && dir=="read")
	    {
		release_read(f,c,0,dp,6)
	    }
	    if (!port_const[f,c,dp] && dir=="write")
	    {
	        printf space(6) "lprintf(\"Switched %s\\n\");\n",port_sel[tf,t,dp]
	    	acquire_write(f,c,0,dp,6,
				"continue;",
				portsize(f,c,dp))
	    	release_write(f,c,0,dp,6,
				variantport(p,tf,ti,"i") "." dp,
				unitportsize(f,c,dp),"",
				unittypesize(f,c,dp))
	    }
	    if (gen) print space(6) "break;"
	}
	if (gen) print space(2) "}}"

	if (!is_node_type(c,n,"unfold")) continue

	dp=p "_" typesels[tf,ti]
	# release selectors

    	if (port_const[c,n,p] && dir=="read")
		release_read(f,c,0,dp,2)

	if (!port_const[c,n,p] && dir=="write") for (z=0; z<portsize(f,c,dp); z++) {
		release_write(f,c,0,dp,2,
			"&" variantsel(p,tf,ti,z),
			unittypesize(f,c,dp),"",
			unittypesize(f,c,dp))
	}
	continue
    }
    # if no - use it for de-activation (do a release)
    if (!is_node_type(f,c,"strict") && future_use(k,c,p)) continue

    if (port_const[c,n,p] && dir=="read")
	    release_read(f,c,n,p,2)

    if (!port_const[c,n,p] && dir=="write") {
	    if (noutfif[f,c,n]==1 && node_ret[c,n]!="void") {
	    	    release_write(f,c,n,p,2,p,"ret*" unittypesize(c,n,p))
		    return
	    }
	    release_write(f,c,n,p,2,p,unitportsize(c,n,p))
    }
  }
}
	

func unitsize(s) { return "sizeof(" s ")" }

func unittypesize(c,n,p) { return unitsize(port_type[c,n,p]) }
func unitportsize(c,n,p) { return "((" portsize(c,n,p) ")*" unittypesize(c,n,p) ")" }

func acquire_read(f,c,n,p,s,P,ret,PS) {
if (1) return
    if (!PS) PS=portsize(c,n,p)
    print space(s) "{struct mq_attr attr;"
    printf space(s+1) "sys_pexit(mq_getattr(%s,&attr),NULL);\n",
    			"args->inpq[" fifoidx(f,c,p) "]"
    printf space(s+1) "if (attr.mq_curmsgs < %s) {\n",
    			 PS
    print space(2*s) "lprintf(\"" c ":  ?could not read %d on port %d: only %d in mq\\n\",",
    			PS ",",
    			fifoidx(f,c,p) ",",
			"attr.mq_curmsgs",
			");"
    print space(2*s) ret
    print space(s) "}}"
}
func do_read(f,c,n,p,s,P,S) {
    print space(s) "{unsigned prio; ssize_t r=1,rd;"
    printf space(s+1) "for (rd=0; r > 0 && rd < %s; rd+=r) {\n", S
    printf space(2*s) "sys_pexit(r=mq_receive(args->inpq[%s],&((char*)%s)[rd],%s-rd,&prio),NULL);\n",
    			fifoidx(f,c,p), P, S
    print space(2*s) "lprintf(\"" c ":read %d (%d) on port %d\\n\",",
    			"r,rd,",
    			fifoidx(f,c,p),
			");"
    print space(s+1) "}"
    print space(s+1) "lprintf(\"" c ":Read %d on port %d\\n\",",
    			"rd,",
    			fifoidx(f,c,p),
			");"
    print space(s) "}"
}

func release_read(f,c,n,p,s, S) { }

func acquire_write(f,c,n,p,s,ret,PS,handle) {
if (1) return
    if (!PS) PS=portsize(c,n,p)
    print space(s) "{struct mq_attr attr;"
    printf space(s+1) "sys_pexit(mq_getattr(%s,&attr),NULL);\n",
    			(handle ? handle : "args->outq[" fifoidx(f,c,p) "]")
    printf space(s+1) "if (attr.mq_curmsgs + %s > attr.mq_maxmsg) {\n",
    			PS
    print space(2*s) "lprintf(\"" c ":  ?could not write %d on port %d: only %d in mq\\n\",",
    			PS ",",
    			fifoidx(f,c,p) ",",
			"attr.mq_maxmsg - attr.mq_curmsgs",
			");"
    print space(2*s) ret
    print space(s+1) "}"
    print space(s+1) "lprintf(\"" c ":  Can write %d on port %d\\n\",",
    			PS ",",
    			fifoidx(f,c,p),
			");"
    print space(s) "}"
}
func release_write(f,c,n,p,s,P,S,handle,TS) {
    if (!TS) TS=unittypesize(c,n,p)
    print space(s) "{unsigned prio=0; ssize_t wr;"
    printf space(s+1) "for(wr=0; wr < %s; wr+=%s) {\n",
    			S, TS
    printf space(2*s) "sys_pexit(mq_send(%s,&((char*)%s)[wr],%s,prio),NULL);\n",
    			(handle ? handle : "args->outq[" fifoidx(f,c,p) "]"),
			P, TS
    print space(s+1) "}"
    print space(s+1) "lprintf(\"" c ":Written %d on port %d\\n\",",
    			S ",",
    			fifoidx(f,c,p),
			");"
    print space(s) "}"
}

@include pragmas.awk
@include gen-shell.awk
@include gxf-parser.awk
@include types-parser.awk
