/****************************************************************
This file is part of LIME.

LIME is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

LIME is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LIME.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/
#include UNIX
#include UNIX_stdio

/* Both of these structs are defined externally */
struct source_state;
struct edge;

void ctor(source_state_t outs[1]) 
{
	printf("called constructor for %s\n",__FILE__);
	outs->count=0;
	outs->size=BUF_SIZE;
}

void process(buf,inst,outs)
	unsigned int buf[restrict BUF_SIZE];
	const source_state_t inst[1];	// no restrict here: shared with outs
	source_state_t outs[1];	// no restrict here: shared with inst
{
	unsigned int i;

	for(i=0; i< BUF_SIZE; i++)
	{
		buf[i] = inst->count;
		outs->count++;
		/* Wrap on signed 24 bits value */
		outs->count &= 0x7fffff;
	}
}

void dtor(const source_state_t inst[1]) {
	printf("called destructor for %s\n",__FILE__);
}

/* Specify local connectity in this module so that
 * the tool can generate complete shells without
 * requiring additional GXF specification:
 *
 * ctor:outs -> process:inst
 * process:outs -> process:inst
 * process:outs -> dtor:inst
 */
#if defined(__LIME__)
struct edge edges[] = {
    {.type="init_state",
     .from={.node=ctor,.port="outs"},
     .to={.node=process,.port="inst"},
    },
    {.type="state",
     .from={.node=process,.port="outs"},
     .to={.node=process,.port="inst"},
    },
    {.type="deinit_state",
     .from={.node=process,.port="outs"},
     .to={.node=dtor,.port="inst"},
    }
};
unsigned int number_of_edges=lengthof(edges);
#endif

/* No schedule array is required here, as the edges array 
 * can be used to re-construct the correct ordering
 */

#if 0
void (*sched[])()={
    ctor, 
    process, 
    dtor 
};
#endif

