##################################################################
# This AWK script contains code shared by all back-ends in LIME
#
#Copyright 2008 Pjotr Kourzanov
#
#This file is part of LIME.
#
#LIME is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License version 2
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#LIME is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with LIME.  If not, see <http://www.gnu.org/licenses/>.
##################################################################
# generic helper functions for generators
func variantport(p,tf,ti,i) { return p "[" i "]." typeunis[tf,ti] }

func variantsel(p,tf,ti,i) { return p "[" i "]." typesels[tf,ti] }

func dump_prototypes(f,c, k,n,j)
{
    for (k=1; k<=nnodes[c]; k++)
    {
    	n=nodes[c,k]
    	printf "extern %s %s(",node_ret[c,n],scpe(c,n)
    	for (j=1; j<=nports[c,n]; j++) {
	    p=ports[c,n,j]
	    printf "%s%s",(j>1 ? "," : ""), get_portdecl(c,n,p)
    	}
    	print ");"
    }
}

# fold port name to the incoming port name
func fold_portvar(f,c,n,p) {
    if (peer_port[c,n,p])
    	return (port_const[c,n,p] ? p : peer_port[c,n,p])

    if (peer_port[f,c,p])
    	return (port_const[f,c,p] ? p : peer_port[f,c,p])

    if (init_port[c,n,p])
    	return (port_const[c,n,p] ? p : init_port[c,n,p])

    if (init_port[f,c,p])
    	return (port_const[f,c,p] ? p : init_port[f,c,p])

    if (deinit_port[c,n,p])
    	return (port_const[c,n,p] ? p : deinit_port[c,n,p])

    if (deinit_port[f,c,p])
    	return (port_const[f,c,p] ? p : deinit_port[f,c,p])

}

func dump_portvar(f,c,n,p,pr,sf) {
    # is this a shmem (e.g., state) port?
    # if yes, the var is a pointer to it...
    if (!is_state(f,c,n,p)) return p

    if (!sf) sf="->"
    return pr "state" sf fold_portvar(f,c,n,p)
}

func dump_ports(f,c,n, j,p)
{
    print space(2) "//",ninports[c,n],"input ports:"
    for (j=1; j<=ninports[c,n];j++) {
    	p=inports[c,n,j]
    	if (p in declared) {
		print space(2) "// already declared:",p
		continue
	}
    	if (is_state(f,c,n,p)) {
		print space(2) "// state port:",p
		continue
	}
    	print space(2) get_portdef(c,n,p) ";"
    	declared[p]=1
    }
    print space(2) "//",noutports[c,n],"output ports:"
    for (j=1; j<=noutports[c,n];j++) {
    	p=outports[c,n,j]
    	if (p in declared) {
		print space(2) "// already declared:",p
		continue
	}
    	if (is_state(f,c,n,p)) {
		print space(2) "// state port:",p
		continue
	}
    	print space(2) get_portdef(c,n,p) ";"
    	declared[p]=1
    }
}

# a way to check whether port p will be used in other
# processing elements futher down the line
func future_use(k,c,p, kk,nn,z) {
    for (kk=k+1; kk<=nnodes[c]; kk++) {
	nn=nodes[c,kk]
	for (z=1; z<=nports[c,nn]; z++) if (ports[c,nn,z]==p)
		return 1
    }
    return 0
}

func is_virtual(f,c,n,p) {
	if ((f,c,p) in delayport_idx) return 1
	if ((f,c,p) in rushport_idx) return 1
	if ((f,c,p) in stateport_idx) return 1
	return 0
}

func is_state(f,c,n,p) {
	if ((f,c,p) in stateport_idx) return 1
	return 0
}

func is_fifo(f,c,n,p) {
	if ((f,c,p) in fifo_idx) return 1
	if ((c,n,p) in fifo_idx) return 1
	return 0
}

func nr_delays(f,c,n) { return ndelayports[f,c,"read"] }
func nr_rushes(f,c,n) { return nrushports[f,c,"write"] }

func nr_states(f,c,n) { return nstateports[f,c,"read"] }
func nr_dstrstates(f,c,n) { return nstateports[f,c,"write"] }

# port numbering is derived from the top-level LIME index
func ninfifos(f,n, r) { r=nfifos[f,n,"read"]; return (r) ? r : ninports[f,n]-nr_states(f,n) }
func noutfifos(f,n, r) { r=nfifos[f,n,"write"]; return (r) ? r : noutports[f,n]-nr_states(f,n) }
func fifoidx(f,c,p, r) { r=fifo_idx[f,c,p]; return (r > 0) ? r-1 : port_idx[f,c,p]-1 }

# TODO data-flow actor can not be a constructor or a destructor
func is_structor(f,n) {
    if ((f,n) in constructors_idx || (f,n) in destructors_idx) 
    {
	if ((f,n) in actors_idx)
		fail("node " n " is both a data-flow actor and a {con,de}structor")
	return 1
    }
    return 0
}

func is_connected(f,n) { return fanout[f,n] || fanin[f,n] }
func is_fullyconnected(f,n) { 
	return nfifos[f,n,"read"]==ninports[f,n]-nr_states(f,n) &&\
	       nfifos[f,n,"write"]==noutports[f,n]-nr_states(f,n) 
}

# TODO: real analysis
func max(a,b) { return (a>=b) ? a : b }

func dnode(n) { return gensub(/\$/,"_","g",n) }

#func comp_fid(pr,n) { return pr "_" n "_fid" }
#func comp_tid(pr,n) { return pr "_" n "_tid" }
#func comp_core(pr,n) { return pr "_" n "_core" }
func comp_fid(n) { return n "_fid" }
func comp_tid(n) { return n "_tid" }
func comp_core(n) { return n "_core" }

func comp_state(pr) { return "struct _" pr "_state" }
func comp_shell(n,pr,sep) { return (pr ? pr (sep ? sep : "_") : "") n "_shell" }
func scpe(c,n) { return c "_" n }

func genedges_c99(f, i,fr,to,pr,fn,fp,tn,tp) {
    pr=gensub(/\.[a-z]+$/,"","g",f)
    print "#include <radio.h>"		
    print "struct edge edges[]={"
    for (i=1; i<=nedges[f]; i++) {
	split(edge_from[f,i],fr,SUBSEP); fn=fr[1]; fp=fr[2]
	split(edge_to[f,i],to,SUBSEP); tn=to[1]; tp=to[2]
	if (is_state(f,fn,"",fp)) continue
	printf space(2) "{ .from={.node=%s,.port=%s /*%s*/}, .to={.node=%s,.port=%s /*%s*/} }%s\n",
		comp_fid(fn),fifoidx(f,fn,fp),fp,
		comp_fid(tn),fifoidx(f,tn,tp),tp,
		(i<nedges[f] ? "," : "")
    }
    print "};"
    print "unsigned int number_of_edges = sizeof edges/sizeof edges[0];"
}
func genedges_old(f, i,fr,to,pr,fn,fp,tn,tp) {
    pr=gensub(/\.[a-z]+$/,"","g",f)
    printf space(2) "setf(nr_of_edges,%u),\n", nedges[f] #TODO
    print space(2) "setf(edge_list, scpe("
    for (i=1; i<=nedges[f]; i++) {
	split(edge_from[f,i],fr,SUBSEP); fn=fr[1]; fp=fr[2]
	split(edge_to[f,i],to,SUBSEP); tn=to[1]; tp=to[2]
	if (is_virtual(f,fn,"",fp)) continue
	printf space(4) "seta(%u, scpe(\n"\
		     "        setf(from, scpe(setf(comp_id,%s),setf(port_id,%u))),\n"\
		     "        setf(to,   scpe(setf(comp_id,%s),setf(port_id,%u))))),\n",
		i-1,
		comp_fid(fn),fifoidx(f,fn,fp),
		comp_fid(tn),fifoidx(f,tn,tp)
    }
    print space(2) ")),"
}
func comp_type(i) {
 	if (comptype[i] == "actor")
		return "TASK"
 	if (comptype[i] == "fifo")
		return "FIFO"
	fail("unknown type " comptype[i])
}
func comp_id(i) {
 	if (comptype[i] == "actor")
		return rname "_" comp_fid(comp[i])
 	if (comptype[i] == "fifo")
		return comp[i]
	fail("unknown type " comptype[i])
}
func put_comp_id(i, t,id) {
	if (!t) t=comp_type(i)
	if (!id) id=comp_id(i)
	return sprintf("setf(comp_id, scpe(setf(s, scpe(\n"\
	   "                               setf(type, %s),\n"\
	   "                               setf(id, %s)))))",
			t, id)
	   	
}
func genedges(f, i,fr,to,pr,fn,fp,tn,tp) {
    pr=gensub(/\.[a-z]+$/,"","g",f)
    printf space(2) "setf(nr_of_edges,%u),\n", 2*nedges[f] #TODO
    print space(2) "setf(edge_list, scpe("
    for (i=1; i<=nedges[f]; i++) {
	split(edge_from[f,i],fr,SUBSEP); fn=fr[1]; fp=fr[2]
	split(edge_to[f,i],to,SUBSEP); tn=to[1]; tp=to[2]
	if (is_virtual(f,fn,"",fp)) continue
	printf space(4) "seta(%u, scpe(\n"\
		     "        setf(from, scpe(%s,setf(port_id,%u))),\n"\
		     "        setf(to,   scpe(%s,setf(port_id,%u))))),\n",
		2*(i-1),
		put_comp_id(0,"TASK",comp_fid(fn)),fifoidx(f,fn,fp),
		put_comp_id(0,"FIFO",edge_id[i]),0
	printf space(4) "seta(%u, scpe(\n"\
		     "        setf(from, scpe(%s,setf(port_id,%u))),\n"\
		     "        setf(to,   scpe(%s,setf(port_id,%u))))),\n",
		2*(i-1)+1,
		put_comp_id(0,"FIFO",edge_id[i]),0,
		put_comp_id(0,"TASK",comp_fid(tn)),fifoidx(f,tn,tp)
    }
    print space(2) ")),"
}
