##################################################################
# shell helper to drive the top-level compilation flow in LIME
#
#Copyright 2008 Pjotr Kourzanov
#
#This file is part of LIME.
#
#LIME is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License version 2
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#LIME is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with LIME.  If not, see <http://www.gnu.org/licenses/>.
##################################################################
export SHELL=/bin/bash

sed_paths=(sed gsed)
sed_version=4.0.7
awk_paths=(awk mawk nawk gawk)
awk_version=3.1.5
make_paths=(make gmake)
make_version=3.81.0
gcc_paths=(gcc gcc-4.3 gcc-4.2 gcc-4.1 gcc-3.4)
gcc_version=3.3.4

auto_paths=(sed awk make gcc)

s="\${"
e="}"
E="[@]}"

sed=sed
awk=awk
make=make
gcc=gcc

fail() {
	echo "FAILED at" $@ >/dev/stderr
	exit
}

function v2n() {
	$sed -e ':lab;s,^\([()*+0-9]\+\)\.,(\1)*100+,g;t lab'
}

function vex() {
	$awk -vIGNORECASE=1 "/(GNU)?.*$p/ && match(\$0,/\<[0-9.]*\>/)\
		{ n=split(substr(\$0,RSTART,RLENGTH),a,\".\");\
		  for (i=1; i<=3; i++)\
		  	printf \"%s%u\",(i>1 ? \".\" : \"\"),a[i] }" 
}

echo -n using >/dev/stderr
for p in ${auto_paths[@]}; do
	eval path="$s${p}_paths$E"
	eval version="$s${p}_version$e"
	vers=$((`echo $version|v2n`))
	found=
	for pa in $path; do
		l=`which $pa 2>/dev/null`
		[[ -z "$l" ]] && continue
		gnu_version=$((`$l --version | vex | v2n`))
		[[ -z "$gnu_version" ]] && continue
		[[ "$gnu_version" -lt "$vers" ]] && continue
		eval $p=$l
		echo -n " $l ($gnu_version)" >/dev/stderr
		found=1
		break
	done
	[[ -n "$found" ]] || fail ": GNU $p $version not found"
done
echo >/dev/stderr

export domain=`echo $base | $sed -ne 's,.*dom/\(.*\)/bin.*,\1,p'`
export lib=$dom/lib
export inc=$dom/inc
export rbin=$root/bin
export rinc=$root/inc

[[ -z "$JOBS" ]] && {
	if [[ -r /proc/cpuinfo ]]; then
		JOBS=`$awk '/^processor/ {p=$3} END {print (p+1)*2}' /proc/cpuinfo`
	else
		JOBS=2
	fi
	echo JOBS set to $JOBS >/dev/stderr
}

DO_INST=0	
[[ "${1##-i}" != "$1" ]] && DO_INST=1 && shift

DO_DOCS=0	
[[ "${1##-d}" != "$1" ]] && shift && [[ "`which noweb`" != "" ]] && [[ "`which cast`" != "" ]] && DO_DOCS=1

DO_CLEAN=0	
[[ "${1##-c}" != "$1" ]] && DO_CLEAN=1 && shift

[[ "${1##-C}" != "$1" ]] && DO_CLEAN=2 && shift

src=
while [[ -n "$1" && "$1" != "--" ]]; do
	if [[ "${1//.c/}" != "$1" ]]; then
		src="$src $1"
	fi
	if [[ "${1//.nw/}" != "$1" ]]; then
		lit="$lit $1"
	fi
	shift
done
[[ "$1" == "--" ]] && shift
# by default include all domain-specific headers into compilation
syshdr=
for h in $rinc/LIME*.h; do
	[[ -r $h ]] || continue
	syshdr="$syshdr $h"
done
export syshdr
hdr=$syshdr
for h in $inc/LIME*.h $inc/$domain*.h; do
	[[ -r $h ]] || continue
	hdr="$hdr $h"
done
while [[ -n "$1" && "$1" != "--" ]]; do
	hdr="$hdr $1"
	shift
done
[[ "$1" == "--" ]] && shift
graph=
while [[ -n "$1" && "$1" != "--" ]]; do
	graph="$graph $1"
	shift
done

clean() {
	[ -n "$msg" ] && echo cleaning $msg >/dev/stderr
	files=;dirs=
	for x in $@; do [ -f "$x" ] && files="$files $x"; done
	for x in $@; do [ -d "$x" ] && dirs="$dirs $x"; done
	[ -n "$files" ] && rm -f $files
	[ -n "$dirs" ] && rm -rf $dirs
	return 0
}

Opt="-j$JOBS sed=$sed awk=$awk make=$make gcc=$gcc DOMAIN=$domain"
Opt="$Opt r=$root rlib=$rlib rbin=$rbin rinc=$rinc inc=$inc"
Opt=$Opt types="$types" modules="$modules" syshdr="$syshdr"

rmake() {
	opt="$Opt -f $rlib/main.mk"
	msg=$1
	if [ ${msg%%:} != ${msg} ]; then
		shift
		if [ -z "$1" ]; then return; fi
		if [ $DO_CLEAN -gt 0 ]; then
			clean $@
			return
		fi
		echo making $msg >/dev/stderr
		$make $opt $@ || fail $msg
	else
		if [ -z "$1" ]; then return; fi
		if [ $DO_CLEAN -gt 0 ]; then
			clean $@
			return
		fi
		$make $opt $@ 
	fi
		
}
dmake() {
	opt="$Opt dom=$dom lib=$lib -f $lib/main.mk"
	msg=$1
	if [ ${msg%%:} != ${msg} ]; then
		shift
		if [ -z "$1" ]; then return; fi
		if [ $DO_CLEAN -gt 0 ]; then
			clean $@
			return
		fi
		echo making $msg >/dev/stderr
		#echo domain $make sed=$sed graph="$graph" graphs="$graphs" comps="$comps" $opt $@
		$make graph="$graph" graphs="$graphs" comps="$comps" $opt $@ || fail $msg
	else
		if [ -z "$1" ]; then return; fi
		if [ $DO_CLEAN -gt 0 ]; then
			clean $@
			return
		fi
		#echo domain $make sed=$sed graph="$graph" graphs="$graphs" comps="$comps" $opt $@
		$make graph="$graph" graphs="$graphs" comps="$comps" $opt $@ 
	fi
}

#echo pre-compiling for $domain: $litc $hierweb $hierpdf $litp $lith >/dev/stderr
litc=${lit//.nw/.c}
litd="${lit//.nw/.nw.tex} ${src//.c/-gen.nw.tex}"

litp="${lit//.nw/.pdf} ${src//.c/-gen.pdf}"
hierpdf="${lit//.nw/-hier.pdf} ${src//.c/-gen-hier.pdf}"

lith="${lit//.nw/.html} ${src//.c/-gen.html}"
hierweb="${lit//.nw/-hier.png} ${src//.c/-gen-hier.png}"

rmake Literate: $litc

src="$src $litc"

#echo generic compiling for $domain: $src -- $hdr -- $graph "(jobs $JOBS)" >/dev/stderr
if [[ $DO_DOCS == 1 ]]; then
	rmake PDF\&WEB: $litd $hierpdf $litp $hierweb $lith
fi

export imports=${src//.c/.imports}
export modules=${src//.c/.module}
export graphs=${graph//.graph/.gxf}
export comps=${src//.c/}

extra="${src//.c/.depx.h} ${src//.c/.domains}"

# don't clean this stuff
if [ $DO_CLEAN != 1 ]; then
    types=
    for x in ${hdr//.h/.types}; do
	rmake $x: $x ${x/.types/.imports}
	export types="$types${types:+ }$x"
    done
fi

rmake Modules\&Graphs: $imports $modules $graphs ${graph//.graph/.dot} $extra

