##################################################################
# Parse XML without DTD/schema validation
#
#Copyright 2008 Pjotr Kourzanov
#
#This file is part of LIME.
#
#LIME is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License version 2
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#LIME is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with LIME.  If not, see <http://www.gnu.org/licenses/>.
##################################################################
BEGIN { state="" }

!NOTRACKFILES { files[f=(FILENAME=="-" ? "stdin" : FILENAME)]=1 }

{ for (unh=1; $0 && unh<=1; )
  {
    consume("[[:space:]]+")

    if (consume("<([?]xml)") ||
        consume("<(!--)") ||
        consume("<(!DOCTYPE)"))
    {
	pass(begtag(el[1]))
	push(el[1])
    } else if (consume("<([^/>][^>]*)>")) {
	end=0
	if (substr(el[1],length(el[1]))=="/") {
		el[1]=substr(el[1],1,length(el[1])-1)
		if (DEBUG) print "element",el[1] >STDERR
		end=1
	}
	n=split(el[1],e)
	# push before handling attributes
	push(e[1])
	# keep id as an attribute (id needs to
	# be an attribute already)
	attrs=""
	delete handled
	for (i=2; i<=n; i++) {
	    split(e[i],v,"=")
	    if (v[1] == "id") {
	    	v[2]=coll(v[1],v[2])
		attrs=append(attrs,v[1] "='" handle("@" v[1],unquote(v[2])) "'")
		#handled[v[1]]=1
	    }
	    if (v[1] ~ /^xlink/) {
	    	v[2]=coll(v[1],v[2])
		attrs=append(attrs,v[1] "='" handle("@" v[1],unquote(v[2])) "'")
		handled[v[1]]=1
	    }
	}
	pass(begtag(e[1],attrs))
	for (i=2; i<=n; i++) {
		split(e[i],v,"=")
		if (v[1] in handled) continue
	    	v[2]=coll(v[1],v[2])
		v[2]=handle(v[1],unquote(v[2]))
		passln(begtag(v[1]) unquote(v[2]) endtag(v[1]))
	}
	if (end) {
		pop()
		handle(e[1])
		passln(endtag(e[1]))
	}
    } else if (endrule(Tag[lev])) { 
   	if (el[1]) Body[lev]=append(Body[lev],el[1])
	pop()
	Body[lev+1]=handle(Tag[lev+1],Body[lev+1])
	passln(Body[lev+1] endtag(Tag[lev+1]))
    } else if (cdata(Tag[lev])) { 
	Body[lev]=append(Body[lev],el[1])
    } else unh++
  }
}

$0 !~ /^[[:space:]]*$/ { fail("XML:" $0) }

func begtag(t,att) {
	if (t=="?xml") return "<" t
	if (t=="!--") return "<" t
	if (t=="!DOCTYPE") return "<" t
	return "<" t pref(att," ") ">" 
}

func endtag(t,pattern) {
	if (t=="?xml") return (pattern ? "[?]>" : "?>" )
	if (t=="!--") return "-->"
	if (t=="!DOCTYPE") return ">"
	return "</" t ">" 
}

func endrule(t) { return state ~ ("@" t "$") && consume("([^<]*)" endtag(t,1)) }

func cdata(t) { return state ~ ("@" t "$") && consume("([^<]*)") }

func unquote(v) { return gensub(/^"(.*)"$/,"\\1","g",gensub(/^'(.*)'$/,"\\1","g",v)) }

func push(st, body) {
	lev++
	state=state "@" st
	if (DEBUG) print lev,"->" state >STDERR
	Tag[lev]=st
	Body[lev]=body
}
func pop() {
	gsub(/@[^@]+$/,"",state)
	lev--
	if (DEBUG) print lev,"->" state >STDERR
}
func consume(pat, r,c) { 
	r=1
	if (pat) r=match($0,"^" pat,el)
	if (DEBUG) print "'" substr($0,RSTART,RLENGTH) "'","consumed" >STDERR
	if (r && !c) $0=substr($0,RLENGTH+1) 
	return r
}
func pass(s) { if (passthrough==1) printf "%s",s }
func passln(s) { return pass(s "\n") }
func coll(v1,v2, qu) {
		if (match(v2,/^(['"])/,qu)) do {
			if (match(v2,qu[1] "$"))
				break
			v2=append(v2,e[++i])
		} while (1)
		return v2
}

@include common.awk
